/*
 * Copyright 1999-2000, University of Notre Dame.
 * Authors: Jeffrey M. Squyres and Dale Southard
 * 
 * This file is part of the Notre Dame Password Storage and Retrieval
 * (PSR) system.
 * 
 * You should have received a copy of the License Agreement for the Notre
 * Dame PSR system along with the software; see the file LICENSE.  If
 * not, contact Office of Research, University of Notre Dame, Notre Dame,
 * IN 46556.
 * 
 * Permission to modify the code and to distribute modified code is
 * granted, provided the text of this NOTICE is retained, a notice that
 * the code was modified is included with the above COPYRIGHT NOTICE and
 * with the COPYRIGHT NOTICE in the LICENSE file, and that the LICENSE
 * file is distributed with the modified code.
 * 
 * LICENSOR MAKES NO REPRESENTATIONS OR WARRANTIES, EXPRESS OR IMPLIED.
 * By way of example, but not limitation, Licensor MAKES NO
 * REPRESENTATIONS OR WARRANTIES OF MERCHANTABILITY OR FITNESS FOR ANY
 * PARTICULAR PURPOSE OR THAT THE USE OF THE LICENSED SOFTWARE COMPONENTS
 * OR DOCUMENTATION WILL NOT INFRINGE ANY PATENTS, COPYRIGHTS, TRADEMARKS
 * OR OTHER RIGHTS.
 * 
 * Additional copyrights may follow.
 * 
 *
 * $Id: afskerbauth.c,v 1.7 2000/12/20 15:59:30 jsquyres Exp $
 *
 * Provides routines necessary for authenticating into AFS using the
 * kerb libraries available from ftp://ftp.pdc.kth.se/pub/krb/src or
 * Kerberos 5 libraries (MIT from http://web.mit.edu/kerberos/www
 * and Heimdal http://www.h5l.org).
 *
 * Authenticate (const char *ln, const char *pw) Does the actual
 * authentication.  It takes pointers to strings with the desired AFS
 * login name and AFS password.  Returns 0 on success.
 *
 * AuthLife (void) Calculates the life of the token to the hour
 * (rounds down) and returns it as an int.  Returns -1 on failure.
 * 
 * GetNewPAG() gets a new PAG free of tokens.
 * 
 */

#include "psr_config.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <unistd.h>
#include <string.h>
#include <limits.h>
#include <fcntl.h>
#include <time.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/stat.h>

#if HAVE_SYS_IOCCOM_H
#include <sys/ioccom.h>
#endif
#include <sys/ioctl.h>

#ifdef KRB4
#include <krb.h>
#endif
#ifdef KRB5
#include <krb5.h>
#ifndef MIT
#include <kafs.h>
#endif
#endif

#include "psr.h"

/* The Authenticate routine.  For AFS we first get a kerb tkt and then
   use the krb_afslog to get an AFS token Return 0 on success. */

#ifdef KRB4
int 
Authenticate(const char *ln, const char *pw)
{
  char tfile[PATH_MAX + 1];
  char env[PATH_MAX + 11];
  int tf;
  int len;
  krb_principal princ;
  struct passwd *ppw;

  if (krb_parse_name(ln, &princ))
    return (1);
  ppw = getpwnam(ln);
  if (! ppw)
    return (1);
/* JMS Did this to make it work on Solaris 7 / 64 bit -- split
   addition of TKT_ROOT and %u_%u into two snprintfs */
  snprintf(tfile, sizeof(tfile), "%s", TKT_ROOT);
  len = strlen(tfile);
  snprintf(tfile + len, sizeof(tfile) - len, "%u_%u",
	   (unsigned) ppw->pw_uid,
	   (unsigned) (getpid()));
  unlink(tfile);
  tf = open(tfile, O_CREAT | O_EXCL | O_RDWR);
  if (tf < 0) {
    printf("couldn't create file \"%s\".\n", tfile);
    return (1);
  }
  close(tf);
  unlink(tfile);
  snprintf(env, sizeof(env), "KRBTKFILE=%s", tfile);
  if (putenv(env) != 0) {
    printf("putenv failed\n");
    return (1);
  }
  krb_set_tkt_string(tfile);
  if (princ.realm[0] == '\0')
    if (krb_get_lrealm(princ.realm, 1) != KSUCCESS)
      strcpy(princ.realm, KRB_REALM);
  if (krb_get_pw_in_tkt(princ.name, princ.instance, princ.realm,
	       KRB_TICKET_GRANTING_TICKET, princ.realm, DEFAULT_TKT_LIFE,
			pw)) {
    printf("Authentication failed (krb4)\n");
    return (1);
  }
  if (k_hasafs())
    if (krb_afslog(NULL, NULL)) {
      printf("Authentication failed (krb4-afs)\n");
      return (1);
    }
  return (0);
}
#endif

#ifdef KRB5
#ifdef MIT
char *afs_ioctl = "/proc/fs/openafs/afs_ioctl";

struct k_ioblock {
  char    *in;
  char    *out;
  uint16_t in_size;
  uint16_t out_size;
};

struct k_pltoken {
  uint32_t kvno;
  char     key[8];
  uint32_t uid;
  uint32_t start;
  uint32_t end;
};

#define VIOCTL_SYSCALL ((unsigned int) _IOW('C', 1, void *))
#define VIOCTL_FN(id)  ((unsigned int) _IOW('V', (id), struct k_ioblock))
#define CIOCTL_FN(id)  ((unsigned int) _IOW('C', (id), struct k_ioblock))
#define OIOCTL_FN(id)  ((unsigned int) _IOW('O', (id), struct k_ioblock))
#define AIOCTL_FN(id)  ((unsigned int) _IOW('A', (id), struct k_ioblock))

int k_ioctl(long fun, long arg1, long arg2, long arg3, long arg4)
{
  int fd, ret;
  struct {
    long arg4;
    long arg3;
    long arg2;
    long arg1;
    long fun;
  } data;
  fd = open(afs_ioctl, O_RDWR);
  if (fd == -1) {
    return  -1;
  }
  data.fun  = fun;
  data.arg1 = arg1;
  data.arg2 = arg2;
  data.arg3 = arg3;
  data.arg4 = arg4;
  ret = ioctl(fd, VIOCTL_SYSCALL, &data);
  close(fd);
  return ret;
}

int k_hasafs(void)
{
  int fd = open(afs_ioctl,O_RDWR);
  if (fd != -1) 
  {
    close(fd);
    return 1;
  } return 0;
}

int k_setpag(void)
{
  return k_ioctl(21L,0L,0L,0L,0L);
}

/* Determine which cell is the default on this workstation */
int k_wscell(char *cell, size_t length)
{
  struct k_ioblock iob;
  char file[] = "/afs";
  memset(&iob, 0, sizeof(iob));
  iob.in       = file;
  iob.in_size  = strlen(file) + 1;
  iob.out      = cell;
  iob.out_size = length - 1;
  memset(cell, '\0', length);
  return k_ioctl(20L, (long)file, VIOCTL_FN(31), (long) &iob, 0L);
}

/* Stuff a ticket and DES key into the kernel */
static int k_settoken(
  const void          *ticket,
  uint32_t             ticket_size,
  int                  kvno, 
  const unsigned char *key,
  uint32_t             uid,
  uint32_t             start,
  uint32_t             end,
  uint32_t             flags,
  const char          *cell)
{
  char                *buffer;
  struct k_ioblock     iob;
  struct k_pltoken     plain_token;
  uint32_t             size;
  int                  i;

  /* Allocate the input buffer. */
  buffer = malloc(4 + ticket_size
		+ 4 + sizeof(plain_token)
		+ 4 + strlen(cell) + 1);
  if (buffer == NULL) {
    return -1;
  }

  /* Their key, encrypted with our key */
  size = ticket_size;
  memcpy(buffer, &size, 4);
  memcpy(buffer + 4, ticket, size);

  /* Our key, plus housekeeping */
  plain_token.kvno  = kvno;
  memcpy(&plain_token.key, key, 8);
  plain_token.uid   = uid;
  plain_token.start = start;
  plain_token.end   = end;
  if (((end - start) % 2) != 0) {
    plain_token.end--;
  }

  size = sizeof(plain_token);
  memcpy(buffer + 4 + ticket_size, &size, 4);
  memcpy(buffer + 4 + ticket_size + 4, &plain_token, size);

  /* Flags */
  size = flags;
  memcpy(buffer + 4 + ticket_size + 4 + sizeof(plain_token), &size, 4);

  /* The name of the cell */
  memcpy(buffer + 4 + ticket_size + 4 + sizeof(plain_token) + 4, cell, strlen(cell) + 1);

  /* The regular stuff */
  memset(&iob, 0, sizeof(iob));
  iob.in       = buffer;
  iob.in_size  = 4 + ticket_size
	       + 4 + sizeof(plain_token)
               + 4 + strlen(cell) + 1;
  iob.out      = NULL;
  iob.out_size = 0;

  i = k_ioctl(20L, (long) NULL, VIOCTL_FN(3), (long) &iob, 0L);
  free(buffer);
  return i;
}

/* Stuff the ticket and key from a v5 credentials structure into the kernel */
static int k_settoken5(const char *cell, krb5_creds *creds, uid_t uid)
{
  /* Assume that the only 8-byte keys are DES keys, and sanity-check */
  if (creds -> keyblock.length != 8) {
    return -1;
  }
  return k_settoken(
    creds -> ticket.data,
    creds -> ticket.length,
    0x100, /* Magic number, signals OpenAFS 1.2.8 and later 
            * that the ticket is actually a v5 ticket */
    creds -> keyblock.contents, uid,
    creds -> times.starttime,
    creds -> times.endtime, 0, cell);
}
#endif

int Authenticate (const char *name, const char *pass)
{
  struct passwd          *user;
  krb5_context            k5con;
  krb5_principal          k5pri;
  krb5_creds              k5cre;
  krb5_ccache             k5cca;
  krb5_get_init_creds_opt k5opt;
  static char             ccfile[PATH_MAX];
  static char             ccname[PATH_MAX];
#ifdef MIT
  krb5_principal          a5pri;
  krb5_creds             *a5cre;
  static char             acell [PATH_MAX];
  static char              cell [PATH_MAX];
#endif

  if ( krb5_init_context(&k5con)
    || krb5_parse_name  ( k5con,name,&k5pri)
    || (user = getpwnam(name)) == NULL
  ) return(1);

  snprintf(ccfile,sizeof(ccfile),
    "/tmp/krb5cc_%u_%lu",
    (     unsigned int) user -> pw_uid,
    (long unsigned int) getpid());
  unlink(ccfile);

  if ( krb5_cc_set_default_name(k5con,ccfile)
    || krb5_cc_default         (k5con,&k5cca)
  ) return(1);

  krb5_get_init_creds_opt_init(&k5opt);

  if (krb5_get_init_creds_password(
        k5con,
       &k5cre,
        k5pri,
        pass ,  krb5_prompter_posix,
        NULL , (krb5_deltat) 0,
        NULL ,
       &k5opt ))
  {
    printf("Authentication failed (krb5)\n");
    return(1);
  }

  snprintf(ccname,sizeof(ccname),"%s:%s",
    krb5_cc_get_type(k5con,k5cca),
    krb5_cc_get_name(k5con,k5cca));

  if ( krb5_cc_initialize(k5con,k5cca, k5cre.client)
    || krb5_cc_store_cred(k5con,k5cca,&k5cre)
    || setenv("KRB5CCNAME",ccname,1))
  {
    printf("Can't create credential cache\n"); 
    return(1);
  }

  if (k_hasafs())
  {
#ifndef MIT
    if (krb5_afslog(k5con,k5cca,NULL,NULL))
#else
    if (k_wscell(cell,PATH_MAX))
    {
      printf("Can't get AFS cell\n");
      return(1);
    }
    strcpy(acell,"afs/");
    strcat(acell,cell);
    if (krb5_parse_name(k5con,acell,&a5pri))
    {
      printf("Can't parse AFS name\n");
      return(1);
    }
    k5cre.server           = a5pri;
    k5cre.keyblock.enctype = ENCTYPE_DES_CBC_MD5;
    if ( krb5_get_credentials(k5con,0,k5cca,&k5cre,&a5cre)
      || k_settoken5(cell,a5cre,user -> pw_uid))
#endif
    {
      printf("Authentication failed (krb5-afs)\n");
      return (1);
    }
  }

  if (chown(ccfile, user->pw_uid, -1) != 0) {
    printf("Can't change owner of \"%s\" to the uid \"%u\"\n",
            ccfile, user->pw_uid);
  }

#ifdef MIT
  krb5_free_creds    (k5con,a5cre);
  krb5_free_principal(k5con,a5pri);
#endif
  krb5_free_principal(k5con,k5pri);
  krb5_free_context  (k5con);

  return(0);
}
#endif

/* The AuthLife routine. Mostly lifted from klist. */

int AuthLife(void)
{
#ifdef MIT
  struct k_ioblock  iob;
  struct k_pltoken  tok;
#else
  struct ViceIoctl  iob;
  struct ClearToken tok;
#endif
  int32_t           size_secret_tok;
  int32_t           size_public_tok;
  unsigned char     t[PATH_MAX];
  unsigned char    *r = t;
  u_int32_t         i = 0;

  iob.in       = (char *)&i ;
  iob.in_size  = sizeof  (i);
  iob.out      = (char *) t ;
  iob.out_size = sizeof  (t);

#ifndef MIT
  if (k_pioctl(NULL, VIOCGETTOK, &iob, 0) < 0)
#else
  if (k_ioctl(20L, (long) NULL, VIOCTL_FN(8), (long) &iob, 0L) < 0)
#endif
  {
    return(0);
  }

  /* Figure out how big secret token is and skip past it */
  memcpy(&size_secret_tok,r,sizeof(size_secret_tok));
  r +=    size_secret_tok + sizeof(size_secret_tok);

  /* Copy the public token into tok */
  memcpy(&size_public_tok,r,sizeof(size_public_tok));
  r +=                      sizeof(size_public_tok);
  memcpy(&tok, r, size_public_tok);

#ifndef MIT
  return ((tok.EndTimestamp - time(NULL))/3600);
#else
  return ((tok.end          - time(NULL))/3600);
#endif
}

/* GetNewPAG() -- gets a new PAG */

void 
GetNewPAG()
{
  if (k_hasafs())
    k_setpag();
}
